/* findproteins.cpp
**
** MoCha: Molecular Characterization
**
** Find and sort by score the proteins that directly or indirectly interact with 
** all the proteins specified in the input, within a pathway length with links
** with at least a minimum score and within a specific organism or any 
** organism. The algorithm uses a set of files derived from the STRING database
** available at (www.string-db.org).
**
** Copyright (C) 2016 Jennifer Hammelman and Daniel Lobo
**               2020 Lobo Lab - https://lobolab.umbc.edu
**
** Version Apr 2020 - Updated to use String v11.
**
** License: GNU GENERAL PUBLIC LICENSE, Version 3
*/

#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <iostream>
#include <map>
#include <unordered_map>
#include <unordered_set>
#include <string>
#include <vector>
#include <algorithm>
#include <cmath>

#define NUM_REQ_ARGS 14
#define MIN_SCORE_ARG (NUM_REQ_ARGS - 5)
#define MAX_DEPTH_ARG (NUM_REQ_ARGS - 4)
#define ORG_ARG (NUM_REQ_ARGS - 3)
#define SCORETYPE_ARG (NUM_REQ_ARGS - 2)
#define PROTEIN1_ARG (NUM_REQ_ARGS - 1)

// Output the usage instructions.
void printUsage() {
  std::cerr << "Usage: findproteins [-h] proteins_file_sorted "
               "proteins_file_sorted_offsets proteins_file_sorted3 "
               "proteins_file_sorted3_offsets links_file_sorted "
               "links_file_sorted_offsets organisms_file_sorted "
               "organisms_file_sorted_offsets min_score max_depth "
               "organism_NCBI_taxonomic_id score_type protein_1 "
               "protein_2 ... protein_n" 
            << std::endl;
}

// Check that the arguments passed are correct.
void checkArgs(int argc, char *argv[]) {
  if (argc > 1 && strcmp(argv[1], "-h") == 0) {
    printUsage();
    std::exit(0);
  } 
  else if (argc < NUM_REQ_ARGS) {
    printUsage();
    std::exit(1);
  }
}

// Open a file.
FILE *openFile(const char *fileName, const char *mode) {
  FILE *file = fopen(fileName, mode);
  if (file == NULL) {
    std::cerr << "Error opening file: " << fileName << std::endl;
    printUsage();
    exit(1);
  }

  return file;
}

// Return the size of an opened binary file.
long binaryFileSize(FILE *file) {
  long prevPos = ftell(file);
  fseek(file, 0, SEEK_END);
  long size = ftell(file);
  fseek(file, prevPos, SEEK_SET);

  return size;
}

// Returns the line offset stored in the offsets file.
long getOffset(FILE *offsetsFile, long line) {
  fseek(offsetsFile, line*sizeof(long), SEEK_SET);
  long offset;
  if (fread(&offset, sizeof(long), 1, offsetsFile) != 1)
    std::cerr << "Error: failed to read offset line " << line << std::endl;
  return offset;
}

// Generic template not implemented
template<typename T>
T getLineElement(FILE *file, FILE *offsetsFile, long line, const char *format, 
                 bool tolower = false) = delete;

// Gets an specific element with a specific format from a file line.
template<>
int getLineElement<int>(FILE *file, FILE *offsetsFile, long line, 
                        const char *format, bool tolower){
  long offset = getOffset(offsetsFile, line);
  fseek(file, offset, SEEK_SET);
  int element;
  if (fscanf(file, format, &element) != 1)
    std::cerr << "Error: failed to scan element in line " << line
              << std::endl;
              
  return element;
}

// Gets an specific element with a specific format from a file line.
// Special case for string.
template<>
std::string getLineElement<std::string>(FILE *file, FILE *offsetsFile, 
                                        long line, const char *format, 
                                        bool tolower) {
  long offset = getOffset(offsetsFile, line);
  fseek(file, offset, SEEK_SET);
  char text[64];
  if (fscanf(file, format, text) != 1)
    std::cerr << "Error: failed to scan string element in line " 
              << line << std::endl;
  if (tolower)
    for (int i = 0; text[i] != '\0'; ++i)
      if (text[i] >= 'A' && text[i] <= 'Z')
        text[i] -= ('Z' - 'z');

  return text;
}

// Gets the first protein in a link from a specific file line
std::string getLinkProtein1(FILE *file, FILE *offsetsFile, long line) {
 return getLineElement<std::string>(file, offsetsFile, line, "%*d.%s");
}

// Gets the second protein in a link from a specific file line
std::string getLinkProtein2(FILE *file, FILE *offsetsFile, long line) {
 return getLineElement<std::string>(file, offsetsFile, line, "%*d.%*s %*d.%s");
}

// Gets the combined score in a link from a specific file line
int getLinkScore(FILE *file, FILE *offsetsFile, long line, 
                 const char *scoreFormat) {
 return getLineElement<int>(file, offsetsFile, line, scoreFormat);
}

// Gets the protein id from a specific file line
std::string getProteinId(FILE *file, FILE *offsetsFile, long line) {
 return getLineElement<std::string>(file, offsetsFile, line, "%*d %s");
}

// Gets the protein common name from a specific file line
std::string getProteinName(FILE *file, FILE *offsetsFile, long line) {
 return getLineElement<std::string>(file, offsetsFile, line, "%*d %*s %s");
}

// Gets the organism taxon_id from a specific file line
int getOrganismId(FILE *file, FILE *offsetsFile, long line) {
 return getLineElement<int>(file, offsetsFile, line, "%d");
}

// Gets the organism name from a specific file line
std::string getOrganismName(FILE *file, FILE *offsetsFile, long line) {
 return getLineElement<std::string>(file, offsetsFile, line, "%*d %*s %[^\t]s");
}

// Find all the lines (first and last) containing the element.
// Based on a double deferred binary search.
template<typename T>
bool findElementLines(FILE *file, FILE *offsetsFile, T element,
                      long *firstLine, long *lastLine, const char *format,
                      bool tolower = false) {
  long minLine = *firstLine;
  long maxLine = *lastLine;

  // find the first line with the element
  long midLine;
  while (minLine < maxLine) {
    midLine = minLine + ((maxLine - minLine) / 2);
    if (getLineElement<T>(file, offsetsFile, midLine, format, tolower) 
        < element)
      minLine = midLine + 1;
    else
      maxLine = midLine;
  }

  bool found = (maxLine == minLine) && 
    (getLineElement<T>(file, offsetsFile, minLine, format, tolower) == element);

  if(found) {
    *firstLine = minLine;
    // find the last line with the element;
    maxLine = *lastLine;
    while (maxLine > minLine) {
      midLine = maxLine - ((maxLine - minLine) / 2);
      if (getLineElement<T>(file, offsetsFile, midLine, format, tolower) 
          > element)
        maxLine = midLine - 1;
      else
        minLine = midLine;
    }

    *lastLine = maxLine;
  }

  return found;
}

// Find all the lines (first and last) containing an organism.
bool findOrgLines(FILE *file, FILE *offsetsFile, int organism,
                  long *firstLine, long *lastLine) {
  return findElementLines(file, offsetsFile, organism, firstLine, lastLine, 
                          "%d");
}

// Find all the lines (first and last) containing a specific first protein in 
// the links file.
bool findProtein1Lines(FILE *file, FILE *offsetsFile, std::string protein, 
                      long *firstLine, long *lastLine) {
  return findElementLines(file, offsetsFile, protein, firstLine, lastLine, 
                          "%*d.%s");
}

// Find all the lines (first and last) containing a specific protein id in 
// the proteins file.
bool findProteinIdLines(FILE *file, FILE *offsetsFile, std::string proteinId, 
                        long *firstLine, long *lastLine) {
  return findElementLines(file, offsetsFile, proteinId, firstLine, lastLine, 
                          "%*d %s");
}

// Find all the lines (first and last) containing a specific protein name in 
// the proteins file (sort3). Ignores protein letter case.
bool findProteinNameLines(FILE *file, FILE *offsetsFile, 
                          std::string proteinName, 
                          long *firstLine, long *lastLine) {
  std::transform(proteinName.begin(), proteinName.end(), proteinName.begin(), 
                 ::tolower);
  return findElementLines(file, offsetsFile, proteinName, firstLine, lastLine, 
                          "%*d %*s %s", true);
}

// Store the details of a protein link.
struct ProteinLink{
  std::string protein2; // the second protein in the link
  int depth; // depth of the link
  long score; // combined score of the link
  
  ProteinLink() {};
  ProteinLink(const std::string &prot2, int dep, int scor) 
    : protein2(prot2), depth(dep), score(scor) {};
};

// Add recursevily to proteinsMap all the proteins linked to protein with a 
// minimum score and within a specific depth. Shortest path to protein has
// preference over better score.
void addProteinLinks(FILE *file, FILE *offsetsFile, long orgFirstLine, 
                 long orgLastLine, std::string protein, int minScore, int depth,
                 std::unordered_map<std::string, ProteinLink> &proteinsMap,
                 const char* scoreFormat) {
  long proteinFirstLineLink = orgFirstLine;
  long proteinLastLineLink = orgLastLine;
  if (!findProtein1Lines(file, offsetsFile, protein,
                         &proteinFirstLineLink, &proteinLastLineLink)) {
    std::cerr << "Warning: no links found from protein " << protein 
              << std::endl;
    return;
  }

  std::vector<std::string> newProteins;

  for (long line = proteinFirstLineLink; line <= proteinLastLineLink; ++line) {
    std::string linkedProtein = getLinkProtein2(file, offsetsFile, line);
    if (proteinsMap.count(linkedProtein) == 0) {
      int score = getLinkScore(file, offsetsFile, line, scoreFormat);
      if (score >= minScore) {
        proteinsMap[linkedProtein] = ProteinLink(protein, depth, score);
        if (depth > 1)
          newProteins.push_back(linkedProtein);
      }
    }
  }

  int numNewProteins = newProteins.size();
  for (int i=0; i < numNewProteins; ++i)
    addProteinLinks(file, offsetsFile, orgFirstLine, orgLastLine, 
                    newProteins[i], minScore, depth-1, proteinsMap, 
                    scoreFormat);
}

// Remove from proteinsOut all the proteins not present in proteinsMapIn
void intersectProteins(
    const std::unordered_map<std::string, ProteinLink> &proteinsMapIn,
          std::unordered_set<std::string> &proteinsOut) {
  std::vector<std::string> proteinsToRemove;
  for (auto it = proteinsOut.begin(); it != proteinsOut.end(); ++it) {
    if (proteinsMapIn.count(*it) == 0) { // Delete not shared protein
      proteinsToRemove.push_back(*it);
    }
  }
  
  int numProteinsToRemove = proteinsToRemove.size();
  for (int i=0; i < numProteinsToRemove; ++i)
    proteinsOut.erase(proteinsToRemove[i]);
}

// Get the protein common name of a protein id.
std::string getProteinName(FILE *file, FILE *offsetsFile, long orgFirstLine, 
                           long orgLastLine, std::string proteinId) {
  long proteinIdFirstLine = orgFirstLine;
  long proteinIdLastLine = orgLastLine;
  bool found = findProteinIdLines(file, offsetsFile, proteinId, 
                                  &proteinIdFirstLine, &proteinIdLastLine);
  if (!found) {
    //std::cout << "Warning: protein common name for " << proteinId
    //          << " not found!" << std::endl;
    return proteinId;
  } 
  else {
    if (proteinIdFirstLine != proteinIdLastLine)
      std::cout << "Warning: multiple protein common names found for " 
                << proteinId << std::endl;
    return getProteinName(file, offsetsFile, proteinIdFirstLine);
  }
}

// Get the protein id of a protein common name (ignore letter case).
bool getProteinId(FILE *file, FILE *offsetsFile, long orgFirstLine, 
                  long orgLastLine, std::string proteinName,
                  std::string &proteinId) {
  long proteinNameFirstLine = orgFirstLine;
  long proteinNameLastLine = orgLastLine;
  bool found = findProteinNameLines(file, offsetsFile, proteinName, 
                                   &proteinNameFirstLine, &proteinNameLastLine);
  if (found) {
    if (proteinNameFirstLine != proteinNameLastLine)
      std::cout << "Warning: multiple protein ids found for " 
                << proteinName << std::endl;
    proteinId = getProteinId(file, offsetsFile, proteinNameFirstLine);
  }

  return found;
}

// Returns the average score of the links of the path that ends in protein.
long calcProteinPathwayScore(const std::string &protein, 
    const std::unordered_map<std::string, ProteinLink> &proteinsMap,
    int maxDepth, FILE *proteinsFile, FILE *proteinsOffsetsFile, 
    long orgFirstLineProteins, long orgLastLineProtteins, 
    std::vector<std::string> &pathway, std::vector<int> &linksScores) {
  const ProteinLink &proteinLink = proteinsMap.at(protein);
  long pathwayScore = pow(1000 - proteinLink.score, 2);

  if (proteinLink.depth != maxDepth) {
    long pathwayScore2 = calcProteinPathwayScore(proteinLink.protein2, 
      proteinsMap, maxDepth, proteinsFile, proteinsOffsetsFile, 
      orgFirstLineProteins, orgLastLineProtteins, pathway, linksScores);
    pathwayScore += pathwayScore2;
  }
  
  pathway.push_back(getProteinName(proteinsFile, proteinsOffsetsFile, 
                                   orgFirstLineProteins, orgLastLineProtteins, 
                                   proteinLink.protein2));
  linksScores.push_back(proteinLink.score);

  return pathwayScore;
}

// Store the details of a protein pathways (with common names).
struct ProteinPathways {
  int organism;
  std::string proteinName;
  std::vector<std::vector<std::string> > pathways;
  std::vector<long> pathwaysScores;
  std::vector<std::vector<int> > linksScores;
};

// Load the organisms to search in
std::map<int, std::string> loadInputOrganisms(int inputOrganism, 
    FILE *organismsFile, FILE *organismsOffsetsFile, long numOrganismsOffsets) {
  std::map<int, std::string> organisms;
  if (!inputOrganism) { // all
    for (int i = 0; i < numOrganismsOffsets-1; ++i)
      organisms[getOrganismId(organismsFile, organismsOffsetsFile, i)] = 
        getOrganismName(organismsFile, organismsOffsetsFile, i);
  } 
  else { // specific input organism
    // Find the organism line in organisms file
    long orgFirstLineOrganisms = 0;
    long orgLastLineOrganisms = numOrganismsOffsets-1;
    if (!findOrgLines(organismsFile, organismsOffsetsFile, inputOrganism,
                      &orgFirstLineOrganisms, &orgLastLineOrganisms)) {
      std::cerr << "Error: organism " << inputOrganism 
                << " not found in organisms file" << std::endl;
      exit(1);
    }

    organisms[inputOrganism] = getOrganismName(organismsFile, 
      organismsOffsetsFile, orgFirstLineOrganisms);
  }

  return organisms;
}

// Calculate all links for each input protein
std::vector<std::unordered_map<std::string, ProteinLink> > calcProteinLinks(
    FILE *linksFile, FILE *linksOffsetsFile, long orgFirstLineLink, 
    long orgLastLineLink, const std::vector<std::string> &inputProteinsIds,
    int minScore, int maxDepth, const std::vector<std::string> &inputProteins,
    const char *scoreFormat) {
  std::vector<std::unordered_map<std::string, ProteinLink> > 
    proteinMaps(inputProteins.size());
  for (int i = 0; i < inputProteins.size(); ++i) {
    addProteinLinks(linksFile, linksOffsetsFile, orgFirstLineLink, 
                    orgLastLineLink, inputProteinsIds[i], minScore, maxDepth, 
                    proteinMaps[i], scoreFormat);

    std::cout << "  Found " << proteinMaps[i].size() << " links from " << 
      inputProteins[i] << std::endl;
  }

  return proteinMaps;
}


// Calculate all common proteins in the map
std::unordered_set<std::string> calcCommonProteins(const 
    std::vector<std::unordered_map<std::string, ProteinLink> > &proteinMaps,
    int numInputProteins) {
  std::unordered_set<std::string> commonProteins;
  const std::unordered_map<std::string, ProteinLink> &proteinMap = 
                                                                proteinMaps[0];
  for (auto it = proteinMap.begin(); it != proteinMap.end(); ++it)
    commonProteins.insert(it->first);
  for (int i = 1; i < numInputProteins; ++i)
    intersectProteins(proteinMaps[i], commonProteins);

  return commonProteins;
}

// Calculate and add to proteinScores the scores of the proteins
void calcPathwayScores(const std::unordered_set<std::string> &proteins, 
    int organism, FILE *proteinsFile, FILE *proteinsOffsetsFile, 
    long orgFirstLineProteins, long orgLastLineProteins, int maxDepth,
    int numInputProteins, const 
    std::vector<std::unordered_map<std::string, ProteinLink> > &proteinMaps,
    std::multimap<float, ProteinPathways> &proteinScores) {
  for (auto it = proteins.begin(); it != proteins.end(); ++it) {
    ProteinPathways proteinPathways;
    proteinPathways.organism = organism;
    proteinPathways.proteinName = getProteinName(proteinsFile, 
      proteinsOffsetsFile, orgFirstLineProteins, orgLastLineProteins, *it);
    proteinPathways.pathways.reserve(numInputProteins);
    proteinPathways.pathwaysScores.reserve(numInputProteins);

    float avgScore = 0;
    for (int i = 0; i < numInputProteins; ++i) {
      std::vector<std::string> pathway;
      std::vector<int> linksScores;
      long pathwayScore = calcProteinPathwayScore(*it, proteinMaps[i], 
        maxDepth, proteinsFile, proteinsOffsetsFile, orgFirstLineProteins, 
        orgLastLineProteins, pathway, linksScores);
      proteinPathways.pathways.push_back(pathway);
      proteinPathways.pathwaysScores.push_back(pathwayScore);
      proteinPathways.linksScores.push_back(linksScores);
      avgScore += pathwayScore;
    }
    avgScore /= numInputProteins;
    proteinScores.insert(std::make_pair(avgScore, proteinPathways));
  }
}

// Find the first and last protein lines for an organism
bool findOrgLineProteins(FILE *proteinsFile, FILE *proteinsOffsetsFile, 
                         int organism, bool exitIfNotFound,
                         long *orgFirstLineProteins, long *orgLastLineProteins){
  if (findOrgLines(proteinsFile, proteinsOffsetsFile, organism,
                   orgFirstLineProteins, orgLastLineProteins)) {
    return true;
  }
  else if (exitIfNotFound) {
      std::cerr << "Error: organism " << organism  
                << " not found in proteins file" << std::endl;
      exit(1);
  }
  else {
    return false;
  }
}

// Find the first and last link lines for an organism
bool findOrgLineLinks(FILE *linksFile, FILE *linksOffsetsFile, 
                         int organism, bool exitIfNotFound,
                         long *orgFirstLineLink, long *orgLastLineLink){
  if (findOrgLines(linksFile, linksOffsetsFile, organism,
                   orgFirstLineLink, orgLastLineLink)) {
    return true;
  }
  else if (exitIfNotFound) {
      std::cerr << "Error: organism " << organism  
                << " not found in links file" << std::endl;
      exit(1);
  }
  else {
    return false;
  }
}

// Find the ids of a set of proteins, abort if not all ids are found
bool findProteinIds(const std::vector<std::string> &inputProteins, 
    FILE *proteinsSor3File, FILE *proteinsSort3OffsetsFile, 
    long orgFirstLineProteins, long orgLastLineProteins, bool exitIfNotAllFound,
    std::vector<std::string> &inputProteinsIds) {
  bool allFound = true;
  int i = 0;
  while (i < inputProteins.size() && allFound) {
    std::string proteinId;
    allFound = getProteinId(proteinsSor3File, proteinsSort3OffsetsFile, 
                            orgFirstLineProteins, orgLastLineProteins, 
                            inputProteins[i], proteinId);
    if (allFound) {
      inputProteinsIds[i] = proteinId;
      ++i;
    }
  }

  if (!allFound && exitIfNotAllFound) {
      std::cout << "Error: protein id for " << inputProteins[i]
                << " not found!" << std::endl;
      exit(1);    
  }

  return allFound;
}

// Search for all the pathways starting from the input proteins
std::multimap<float, ProteinPathways> searchPathways(FILE *proteinsFile, 
    FILE *proteinsOffsetsFile, long numProteinsOffsets, FILE *proteinsSor3File, 
    FILE *proteinsSort3OffsetsFile, FILE *linksFile, FILE *linksOffsetsFile, 
    long numLinksOffsets, int inputOrganism, int minScore, int maxDepth, 
    const std::map<int, std::string> &organisms, 
    const std::vector<std::string> &inputProteins, const char *scoreFormat) {
  std::multimap<float, ProteinPathways> proteinScores;

  for (auto it = organisms.cbegin(); it != organisms.cend(); ++it) {
    int organism = it->first;

    // Calculate the organism first and last lines in proteins file
    long orgFirstLineProteins = 0;
    long orgLastLineProteins = numProteinsOffsets-1;
    if(!findOrgLineProteins(proteinsFile, proteinsOffsetsFile, organism, 
                            inputOrganism, &orgFirstLineProteins, 
                            &orgLastLineProteins))
      continue; // Continue with next organism

    // Find input proteins ids
    std::vector<std::string> inputProteinsIds(inputProteins.size());
    if(!findProteinIds(inputProteins, proteinsSor3File, 
                       proteinsSort3OffsetsFile, orgFirstLineProteins, 
                       orgLastLineProteins, inputOrganism, inputProteinsIds))
      continue; // Continue with next organism

    // Calculate the organism first and last lines in links file
    long orgFirstLineLink = 0;
    long orgLastLineLink = numLinksOffsets-1;
    if(!findOrgLineLinks(linksFile, linksOffsetsFile, organism, inputOrganism, 
                         &orgFirstLineLink, &orgLastLineLink))
      continue; // Continue with next organism
    

    std::cout << "Searching in organism " << organisms.at(organism) 
              << " (taxon id: " << organism << " num named proteins: "
              << orgLastLineProteins - orgFirstLineProteins << " num links: "
              << orgLastLineLink - orgFirstLineLink << ") ..." << std::endl;
    
    std::vector<std::unordered_map<std::string, ProteinLink> > proteinMaps = 
      calcProteinLinks(linksFile, linksOffsetsFile, orgFirstLineLink, 
                       orgLastLineLink, inputProteinsIds, minScore, maxDepth, 
                       inputProteins, scoreFormat);


    // Find the proteins linked by all the input proteins
    std::unordered_set<std::string> commonProteins = calcCommonProteins(
      proteinMaps, inputProteins.size());

    std::cout << "  Found " << commonProteins.size() << " common proteins"
              << std::endl;
    
    // Calculate average scores and pathways with protein common names
    calcPathwayScores(commonProteins, organism, proteinsFile, 
                      proteinsOffsetsFile, orgFirstLineProteins, 
                      orgLastLineProteins, maxDepth, inputProteins.size(), 
                      proteinMaps, proteinScores);
  }

  return proteinScores;
}

// Print the results
void printResults(const std::multimap<float, ProteinPathways> &proteinScores,
                  const std::map<int, std::string> &organisms,
                  int numInputProteins) {
  std::cout << std::endl << "Common proteins found (total " 
            << proteinScores.size() << "):" << std::endl << std::endl;
  
  int i = 1;
  for (auto it = proteinScores.cbegin(); it != proteinScores.cend(); 
       ++it, ++i) {
    float score = it->first;
    const ProteinPathways &proteinPathways = it->second;
    std::cout << i << ". " << organisms.at(proteinPathways.organism) << " - " 
              << proteinPathways.proteinName << " (average RSS score: " << score 
              << "): " << std::endl;
    
    for (int j=0; j < numInputProteins; ++j) {
      const std::vector<std::string> &pathway = proteinPathways.pathways[j];
      std::cout << "  " << pathway[0];
      for (auto it = pathway.begin()+1; it != pathway.end(); ++it)
        std::cout << "->" << *it;

      std::cout << "->" << proteinPathways.proteinName << " (RSS score: "
                << proteinPathways.pathwaysScores[j] << "  links scores:";
      
      const std::vector<int> &linksScores = proteinPathways.linksScores[j];
      for (auto it = linksScores.begin(); it != linksScores.end(); ++it)
        std::cout << " " << *it;
                
      std::cout << ")" << std::endl;
    }
    std::cout << std::endl;
  }
}

// Print some general information at the beginning of the output
void printHeader(long numOrganismsOffsets, long numProteinsOffsets, 
                 long numLinksOffsets, int inputOrganism, 
                 const std::map<int, std::string> &organisms,
                 const std::vector<std::string> &inputProteins, int minScore,
                 int maxDepth, int scoreType) {
  std::cout << "MoCha: searching in " << numOrganismsOffsets << " organisms, " 
            << numProteinsOffsets << " proteins with common names and "
            << numLinksOffsets << " links." << std::endl << std::endl;
  
  std::cout << "Searching for all the proteins interacting with " 
            << inputProteins[0];
  for (int i = 1; i < inputProteins.size()-1; ++i)
    std::cout << ", " << inputProteins[i];

  if (inputProteins.size() > 1)
    std::cout << " and " << inputProteins[inputProteins.size()-1];

  if (inputOrganism)
    std::cout << " in " << organisms.at(inputOrganism) << " (taxon id " 
              << inputOrganism << ")";
  else
    std::cout << " in any organism";

  std::cout << " in pathways with " << maxDepth << " links maximum and "
            << "with link scores of type " << scoreType << " and equal or "
            << "higher than " << minScore << ". ";

  std::cout << "Pathway scores indicate the residual sum of squares (RSS) "
               "of the link scores (lower is"
               " better confidence). Link scores indicate the STRING confidence"
               " of the proteins interaction (higher is better confidence)."
            << std::endl << std::endl;
}

// Main program
int main(int argc, char *argv[]) {
  // Load arguments and open files
  checkArgs(argc, argv);
  FILE *proteinsFile = openFile(argv[1], "r");
  FILE *proteinsOffsetsFile = openFile(argv[2], "r");
  long numProteinsOffsets = (binaryFileSize(proteinsOffsetsFile) / sizeof(long)) -1;
  FILE *proteinsSor3File = openFile(argv[3], "r");
  FILE *proteinsSort3OffsetsFile = openFile(argv[4], "r");

  FILE *linksFile = openFile(argv[5], "r");
  FILE *linksOffsetsFile = openFile(argv[6], "r");
  long numLinksOffsets = (binaryFileSize(linksOffsetsFile) / sizeof(long)) -1;

  FILE *organismsFile = openFile(argv[7], "r");
  FILE *organismsOffsetsFile = openFile(argv[8], "r");
  long numOrganismsOffsets = (binaryFileSize(organismsOffsetsFile) / sizeof(long)) -1;
  
  int minScore = atoi(argv[MIN_SCORE_ARG]);
  if (minScore < 0)
    minScore = 0;

  int maxDepth = atoi(argv[MAX_DEPTH_ARG]);
  if (maxDepth < 1)
    maxDepth = 1;

  int inputOrganism = atoi(argv[ORG_ARG]); // 0 if no number (all organisms)
  
  int scoreType = atoi(argv[SCORETYPE_ARG]);
  
  if (scoreType < 0)
    scoreType = 0;
  else if (scoreType > 7)
    scoreType = 7;

  char scoreFormat[64] = "%*d.%*s %*d.%*s";
  for (int i = 0; i < scoreType; ++i)
    sprintf(scoreFormat+15+4*i, " %%*d");
  
  sprintf(scoreFormat+15+4*scoreType, " %%d");

  int numInputProteins = argc - PROTEIN1_ARG;
  std::vector<std::string> inputProteins(numInputProteins);
  for (int i = 0; i < numInputProteins; ++i)
    inputProteins[i] = argv[PROTEIN1_ARG+i];
  
  // Load input organisms
  std::map<int, std::string> organisms = loadInputOrganisms(inputOrganism, 
    organismsFile, organismsOffsetsFile, numOrganismsOffsets);

  // Print header text
  printHeader(numOrganismsOffsets, numProteinsOffsets, numLinksOffsets, 
              inputOrganism, organisms, inputProteins, minScore, maxDepth,
              scoreType);

  // Search for all the pathways starting from the input proteins
  std::multimap<float, ProteinPathways> proteinScores = searchPathways(
    proteinsFile, proteinsOffsetsFile, numProteinsOffsets, proteinsSor3File, 
    proteinsSort3OffsetsFile, linksFile, linksOffsetsFile, numLinksOffsets, 
    inputOrganism, minScore, maxDepth, organisms, inputProteins, scoreFormat);


  // Print the found pathways
  printResults(proteinScores, organisms, numInputProteins);


  // Close files
  fclose(linksFile);
  fclose(linksOffsetsFile);

  return 0;
}

